import { Promise } from "./promise";
import { Queue } from "./queue";
import { CancellationError } from "./cancellation_error";

// https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Error
function ResolvedMonitorError(message) {
  this.name = "resolved";
  if ((message !== undefined) && (typeof message !== "string")) {
    throw new TypeError('You must pass a string.');
  }
  this.message = message || "Default Message";
}
ResolvedMonitorError.prototype = new Error();
ResolvedMonitorError.prototype.constructor = ResolvedMonitorError;

function isFunction(x){
  return typeof x === "function";
}

var Monitor = function() {
  var monitor = this,
    promise_list = [],
    promise,
    reject,
    notify,
    resolved;

  if (!(this instanceof Monitor)) {
    return new Monitor();
  }

  function canceller() {
    var len = promise_list.length,
      i;
    for (i = 0; i < len; i += 1) {
      promise_list[i].cancel();
    }
    // Clean it to speed up other canceller run
    promise_list = [];
  }

  promise = new Promise(function(done, fail, progress) {
    reject = function (rejectedReason) {
      if (resolved) {return;}
      monitor.isRejected = true;
      monitor.rejectedReason = rejectedReason ;
      resolved = true;
      canceller();
      return fail(rejectedReason);
    };
    notify = progress;
  }, canceller);

  monitor.cancel = function () {
    if (resolved) {return;}
    resolved = true;
    promise.cancel();
    promise.fail(function (rejectedReason) {
      monitor.isRejected = true;
      monitor.rejectedReason = rejectedReason;
    });
  };
  monitor.then = function () {
    return promise.then.apply(promise, arguments);
  };

  monitor.monitor = function(promise_to_monitor) {
    if (resolved) {
      throw new ResolvedMonitorError();
    }
    var queue = new Queue()
      .push(function () {
        return promise_to_monitor;
      })
      .push(function (fulfillmentValue) {
        // Promise to monitor is fullfilled, remove it from the list
        var len = promise_list.length,
          promise_to_monitor,
          new_promise_list = [],
          i;
        for (i = 0; i < len; i += 1) {
          promise_to_monitor = promise_list[i];
          if (!(promise_to_monitor.isFulfilled ||
              promise_to_monitor.isRejected)) {
            new_promise_list.push(promise_to_monitor);
          }
        }
        promise_list = new_promise_list;
      }, function (rejectedReason) {
        if (rejectedReason instanceof CancellationError) {
          if (!(promise_to_monitor.isFulfilled && promise_to_monitor.isRejected)) {
            // The queue could be cancelled before the first push is run
            promise_to_monitor.cancel();
          }
        }
        reject(rejectedReason);
        throw rejectedReason;
      }, function (notificationValue) {
        notify(notificationValue);
        return notificationValue;
      });

    promise_list.push(queue);

    return this;
  };
};

Monitor.prototype = Object.create(Promise.prototype);
Monitor.prototype.constructor = Monitor;

export { Monitor, ResolvedMonitorError };
